/*
 *  Simon for Uzebox - implementation of the electronic game Simon says 
 *  Version 1.0
 *  Copyright (C) 2011  Hartmut Wendt
 *
 *  This program is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/


#define _Fading	// enables FadeIn & FadeOut between the different screens
#define _Music	// enables Music
//#define _Meine_WII	// enables reworked Gravis Gamepad as input device
#define Music_volume 50	// volume for music
#define SFX_volume 250	// volume for sound effects

#include <stdbool.h>
#include <string.h>
#include <avr/io.h>
#include <stdlib.h>
#include <avr/pgmspace.h>
#include "kernel/uzebox.h"
#include "data/simon_tiles.inc"
#include "data/simon_logo.inc"
#include "data/fonts.pic.inc"
#include "data/patches.inc"
#ifdef _Music
 #include "data/simon_track.inc"
#endif



/* global definitons */
// program modes
#define PM_Intro	0	// program mode intro
#define PM_Simon 	1	// program mode: Simon Says ...
#define PM_Player 	2	// program mode: answer from player
#define PM_Credits	3	// shows credits
#define PM_HoF_view	4	// view hall of fame
#define PM_HoF_edit	5	// edit hall of fame
#define PM_Game_over  	20	// program mode for game over screen


// simon colors
#define red			1
#define blue		2
#define green		3
#define yellow		4


// position simon
#define Simon_X		13
#define Simon_Y 	9

// general parameters
#define max_steps	100


// 8-bit, 255 period LFSR (for generating pseudo-random numbers)
#define PRNG_NEXT() (prng = ((u8)((prng>>1) | ((prng^(prng>>2)^(prng>>3)^(prng>>4))<<7))))





//#define delta_t 1 

#define MAX(x,y) ((x)>(y) ? (x) : (y))


struct EepromBlockStruct ebs;



u8 prng; // Pseudo-random number generator
u8 program_mode;	// program mode (intro, 1 player mode, 2 player mode, ....
int PosX=0, PosY=0;	 // position of cursor
u8 steps[max_steps + 1];
u8 current_step;
int iSCORE;			// Highscore
bool Music_on= true;



/*** function prototypes *****************************************************************/
void init(void);
void set_PM_mode(u8 mode);
void draw_simon(u8 color);
void copy_buf(unsigned char *BUFA, unsigned char *BUFB, unsigned char ucANZ);
void fill_buf(unsigned char *BUFA, unsigned char value, unsigned char ucANZ);
void msg_window(u8 x1, u8 y1, u8 x2, u8 y2);
int get_highscore(u8 entry);
u8 check_highscore(void);
void copy_highsore(u8 entry_from, u8 entry_to);
void clear_highsore(u8 entry);
u8 set_def_highscore(void);
u8 view_highscore_entry(u8 x, u8 y, u8 entry, u8 load_data);
void edit_highscore_entry(u8 entry, u8 cursor_pos, u8 b_mode);
void show_highscore_char(u8 entry, u8 position, u8 cursor_on);
//void fade_out_volume(void);
void simon_end(bool game_winner);


void init(void)
// init program
{
  // init font table
  SetFontTable(fonts);
	

  // init music	
  InitMusicPlayer(patches);
  #ifdef _Music
	 StartSong(popcorn);
  #endif
  SetMasterVolume(Music_volume);

  // load into screen
  set_PM_mode(PM_Intro);

  // init random generator
  prng = 15;

  //Use block 23  
  ebs.id = 23;
  if (!isEepromFormatted())
     return;

  if (EEPROM_ERROR_BLOCK_NOT_FOUND == EepromReadBlock(23,&ebs))
  {
	set_def_highscore();
  }		
}



int main(){
int ibuttons=0,ibuttons_old;
u8 b;

  // init program
  init();        
  // proceed game	
  while(1)
  {
    WaitVsync(3);	  
    // get controller state
    ibuttons_old = ibuttons;
	ibuttons = ReadJoypad(0);
    switch(program_mode)
	{
	  // proceed intro mode	
	  case PM_Intro:

     	if ((BTN_A & ibuttons) && !(ibuttons_old & BTN_A)) {
		  // wait for release of button A 
		  while (BTN_A & ReadJoypad(0)) {
		    prng++; 		
			WaitVsync(1);
		  }
      	  prng = MAX(prng,1);
		  if (PosY == 0) set_PM_mode(PM_Simon);	
		  else if (PosY == 1) set_PM_mode(PM_Credits);	
  		  else if (PosY == 2) {
		    iSCORE = 0; 
		    set_PM_mode(PM_HoF_view);

		  } else {
		    if (Music_on) {
			  Music_on = false;
			  Print(16,21,PSTR("MUSIC OFF "));
			  StopSong();
		    } else {
			  Music_on = true;			  
			  Print(16,21,PSTR("MUSIC ON "));
			  ResumeSong();
			}
		  }
 		}
		// UP button
		if ((ibuttons & BTN_UP) && !(ibuttons_old & BTN_UP)) {

		  // wait for release of button BTN_UP 
		  while (BTN_UP & ReadJoypad(0));

		  PrintChar(15,18+PosY,' '); 	
		  if (PosY) PosY--;	
		  PrintChar(15,18+PosY,'^'); 	
		  

		// DOWN button
		} else if ((ibuttons & BTN_DOWN) && !(ibuttons_old & BTN_DOWN)) {

		  // wait for release of button BTN_DOWN 
		  while (BTN_DOWN & ReadJoypad(0));

		  PrintChar(15,18+PosY,' '); 	
		  if (PosY<3) PosY++;	
		  PrintChar(15,18+PosY,'^'); 	
		  
		}


        break;

	  // simon says	..
	  case PM_Simon:
		Print(1,1,PSTR("SIMON SAYS ...       "));
			
		// calculate next step and increase the count of steps
		steps[current_step++] = 1 + PRNG_NEXT() % 4;
		
		// check maximal possible count of steps	
		if (current_step >= max_steps) {
		   simon_end(true);
		   break;
        }

		// Print actual score
		PrintByte(9,26,current_step,false);

		WaitVsync(20);
		for(b=0; b < current_step; b++) {

		  draw_simon(steps[b]);
		  WaitVsync(15);	
		  draw_simon(0);
		  WaitVsync(15);	
		}	
		WaitVsync(10);
		// reset counter
		b=0;
		// switch to player mode
		set_PM_mode(PM_Player);
        break;


	  // player says...
	  case PM_Player:			
		if (b < current_step) {
			
			#ifdef _Meine_WII
			// button definition for Uzebox with ATART compatible inputs and Gravis Gamepad
			// button B (green) pressed
     		if ((BTN_B & ibuttons) && !(ibuttons_old & BTN_B)) {				
		  		draw_simon(green);
		  		if (green != steps[b]) {
				  simon_end(false);	
				  break;
				} 
				b++;

		  		// wait for release of button B 
		  		while (BTN_B & ReadJoypad(0));
		  		draw_simon(0);		    
			} 

			// button A (yellow) pressed
			if ((BTN_A & ibuttons) && !(ibuttons_old & BTN_A)) { 				
		  		draw_simon(yellow);
		  		if (yellow != steps[b]) {
				  simon_end(false);	
				  break;
				} 
				b++;

		  		// wait for release of button A
		  		while (BTN_A & ReadJoypad(0));
		  		draw_simon(0);
			} 

			// button Y (blue) pressed
			if ((BTN_Y & ibuttons) && !(ibuttons_old & BTN_Y)) {  
		  		draw_simon(blue);
		  		if (blue != steps[b]) {
				  simon_end(false);	
				  break;
				} 
				b++;

		  		// wait for release of button Y 
		  		while (BTN_Y & ReadJoypad(0));
		  		draw_simon(0);
			} 

			// button START (red) pressed
			if ((BTN_START & ibuttons) && !(ibuttons_old & BTN_START)) {  
		  		draw_simon(red);
		  		if (red != steps[b]) {
				  simon_end(false);	
				  break;
				} 
				b++;

		  		// wait for release of button START 
		  		while (BTN_START & ReadJoypad(0));
		  		draw_simon(0);
			}


			#else
			// buttin definition for standard UZEBOX and EUZEBOX with SNES Controller
			// button A (red) pressed
     		if ((BTN_A & ibuttons) && !(ibuttons_old & BTN_A)) {				
		  		draw_simon(red);
		  		if (red != steps[b]) {
				  simon_end(false);	
				  break;
				} 
				b++;

		  		// wait for release of button A 
		  		while (BTN_A & ReadJoypad(0));
		  		draw_simon(0);		    
			} 

			// button B (yellow) pressed
			if ((BTN_B & ibuttons) && !(ibuttons_old & BTN_B)) { 				
		  		draw_simon(yellow);
		  		if (yellow != steps[b]) {
				  simon_end(false);	
				  break;
				} 
				b++;

		  		// wait for release of button B 
		  		while (BTN_B & ReadJoypad(0));
		  		draw_simon(0);
			} 

			// button X (blue) pressed
			if ((BTN_X & ibuttons) && !(ibuttons_old & BTN_X)) {  
		  		draw_simon(blue);
		  		if (blue != steps[b]) {
				  simon_end(false);	
				  break;
				} 
				b++;

		  		// wait for release of button X 
		  		while (BTN_X & ReadJoypad(0));
		  		draw_simon(0);
			} 

			// button Y (green) pressed
			if ((BTN_Y & ibuttons) && !(ibuttons_old & BTN_Y)) {  
		  		draw_simon(green);
		  		if (green != steps[b]) {
				  simon_end(false);	
				  break;
				} 
				b++;

		  		// wait for release of button Y 
		  		while (BTN_Y & ReadJoypad(0));
		  		draw_simon(0);
			}
			#endif

		} else program_mode = PM_Simon;

        break;

	  // show credits screen...
	  case PM_Credits:			
		WaitVsync(120);				
		set_PM_mode(PM_Intro);		
        break;

	  case PM_HoF_view:				
		// time control				
		if (b <= 200) b++;
		if (b == 50) Print(14,24,PSTR("PRESS BUTTON A"));
		if (((ibuttons & BTN_A) && !(ibuttons_old & BTN_A) && (b > 50)) || (b > 200))
		{
		  b = 0;	
		  set_PM_mode(PM_Intro);
		}
		break;

	  case PM_HoF_edit:				
		// cursor blinking
		b++;
		if (b >= 10) b = 0;
		// proceed cursor position with left & right button
		if ((ibuttons & BTN_RIGHT) && !(ibuttons_old & BTN_RIGHT)) {
		  if (PosX < 7) {       	   
		  	show_highscore_char(PosY - 1, PosX, 0); 		 
		    PosX++; 			
          }
		}
		if ((ibuttons & BTN_LEFT) && !(ibuttons_old & BTN_LEFT)) {		 
 		  if (PosX) {
		    show_highscore_char(PosY - 1, PosX, 0); 
		    PosX--; 
          }
		}
		// chose character up & down button
		if ((ibuttons & BTN_UP) && !(ibuttons_old & BTN_UP)) {
		  edit_highscore_entry(PosY,PosX,BTN_UP); 
		}
		else if ((ibuttons & BTN_DOWN) && !(ibuttons_old & BTN_DOWN)) {		 
 		  edit_highscore_entry(PosY,PosX,BTN_DOWN); 
		}     
		// show cursor
		show_highscore_char(PosY - 1, PosX, b > 4);

		// store new entry
		if (ibuttons & BTN_A)   
		{
		  // store new highscore 
		  EepromWriteBlock(&ebs);
		  set_PM_mode(PM_Intro);
		}
		break;

	}

  }
  
} 



void set_PM_mode(u8 mode) {
// set parameters, tiles, background etc for choosed program mode		
	switch (mode)
	{

	  case	PM_Intro:

		// init variables
		PosY = 0;
   		 	    
		//clear screen with fade
		#ifdef _Fading
		 FadeOut(1, true);
		#endif

		ClearVram();

		//Set the tile table to use.
		SetTileTable(LogoTiles);
		WaitVsync(2);

		// draw Simon logo
  		DrawMap(3,1,simon_logo);        

        //Print(4,8,PSTR("VERSION")); 

        Print(14,9,PSTR("VERSION 1.0")); 
        Print(3,11,PSTR("FOLLOW AND REPEAT THE SEQUENCE OF")); 
        Print(3,12,PSTR("LIGHTS AND SOUNDS FOR AS LONG AS")); 
        Print(8,13,PSTR("YOU CAN ... IF YOU CAN!")); 

        Print(15,18,PSTR("^PLAY")); 
        Print(16,19,PSTR("CREDITS")); 
        Print(16,20,PSTR("HALL OF FAME")); 

	    if (Music_on) {
		  Print(16,21,PSTR("MUSIC ON "));
	    } else {
		  Print(16,21,PSTR("MUSIC OFF "));
		}
 

        Print(7,26,PSTR("LICENSED UNDER GNU GPL V3"));

		#ifdef _Fading
		 FadeIn(1, true);
		#endif

		break;
	  

     case	PM_Simon:
	 	// simon says

		// stop music
		StopSong();
		// full volume
        SetMasterVolume(SFX_volume);	


		//clear screen with fade
		#ifdef _Fading
		 FadeOut(1, true);
		#endif

		ClearVram();


		current_step = 0;
		fill_buf(&steps[0],0,100);

		//Set the tile table to use.
		SetTileTable(BGTiles);
   		 
		//Draw Simon
		DrawMap(Simon_X,Simon_Y,simon_all_off); 

		//Draw texts
		Print(1,26,PSTR("SCORE:")); 
		Print(25,26,PSTR("HIGHSCORE:")); 

		// Get and draw actual Highscore
		PrintByte(38,26,get_highscore(0),false);

		#ifdef _Fading
		 FadeIn(1, true);
		#endif

		break;

      case	PM_Player:
	 	// player says
		Print(1,1,PSTR("DO WHAT SIMON SAYS..."));
		break;
	

      case	PM_Credits:
	 	// show credits

		//clear screen with fade
		#ifdef _Fading
		 FadeOut(1, true);
		#endif

		ClearVram();


        Print(4,10,PSTR("COPYRIGHT 2011 BY HARTMUT WENDT"));  
        Print(10,12,PSTR("WWW.HWHARDSOFT.DE.VU"));  


        Print(6,15,PSTR("SOUND TRACK BY CARSTEN KUNZ"));  

		#ifdef _Fading
		 FadeIn(1, true);
		#endif

		break;

      case	PM_HoF_view:
	    // show the hall of fame

    	//clear screen with fade
		#ifdef _Fading
		 FadeOut(1, true);
		#endif

		ClearVram();


	    PosY = check_highscore();
	    if (PosY == 2) copy_highsore(1,2);
	    if (PosY == 1) {
		  copy_highsore(1,2);
		  copy_highsore(0,1);
        }
		clear_highsore(PosY - 1);

		// reset cursor to left position
		PosX = 0;

	  	// Set the font and tiles to use.
		SetTileTable(LogoTiles);
	    ClearVram();
		msg_window(12,8,29,19);	
		Print(15,10,PSTR("HALL OF FAME"));
        view_highscore_entry(14,13,1,!(PosY));
        view_highscore_entry(14,15,2,!(PosY));
        view_highscore_entry(14,17,3,!(PosY));  
		if (PosY)  {
		  mode = PM_HoF_edit;
		  Print(14,3,PSTR("CONGRATULATION"));	
		  Print(14,5,PSTR("NEW HIGHSCORE!"));	
		  Print(13,22,PSTR("ENTER YOUR NAME"));	
		  Print(12,24,PSTR("AND PRESS BUTTON A"));	
		}
		#ifdef _Fading
		  FadeIn(1, true);
		#endif
		break;	  


	}
	program_mode = mode;

}



void draw_simon(u8 color) {
// draw simon with selected colour
  switch (color) {
	case red:
	    DrawMap(Simon_X,Simon_Y,simon_all_off); 
		DrawMap(Simon_X+8,Simon_Y+2,simon_red_on); 
        //Sound
     	TriggerFx(3, 0x7f, true);
		break;

	case green:
		DrawMap(Simon_X,Simon_Y,simon_all_off); 		
		DrawMap(Simon_X,Simon_Y+2,simon_green_on); 
        //Sound
		TriggerFx(5, 0x7f, true);
		break;

	case blue:
	    DrawMap(Simon_X,Simon_Y,simon_all_off); 
		DrawMap(Simon_X+2,Simon_Y,simon_blue_on); 
        //Sound
		TriggerFx(4, 0x7f, true);
		break;

	case yellow:
		DrawMap(Simon_X,Simon_Y,simon_all_off); 
		DrawMap(Simon_X+2,Simon_Y+6,simon_yellow_on); 
        //Sound
     	TriggerFx(6, 0x7f, true);
		break;

	default:	
	  	DrawMap(Simon_X,Simon_Y,simon_all_off); 		
		break;


  }

}


void simon_end(bool game_winner) {
  // you loose the game against Simon

  draw_simon(0); // all colors off
  if (game_winner) Print(1,1,PSTR("YOU ARE THE CHAMPION!   "));
  else Print(1,1,PSTR("SIMON WINS !!!          "));       
  WaitVsync(10);	
  // play loose sound
  TriggerFx(7, 0x7f, true);	
  WaitVsync(90);
  
  iSCORE = current_step - 1;
  if (0 == check_highscore()) set_PM_mode(PM_Intro);	  	
  else set_PM_mode(PM_HoF_view);	

  // set music on if ok
  SetMasterVolume(Music_volume);	
  if (Music_on) ResumeSong();

   
}


void copy_buf(unsigned char *BUFA, unsigned char *BUFB, unsigned char ucANZ)
// copy ucANZ Bytes from buffer BUFA to buffer BUFB
{
 for(;ucANZ>0 ; ucANZ--)
 {
  *(BUFB++) = *(BUFA++);
 }   
}



void fill_buf(unsigned char *BUFA, unsigned char value, unsigned char ucANZ)
// fill ucANZ count of bytes of buffer BUFA with value
{
 for(;ucANZ>0 ; ucANZ--)
 {
  *(BUFA++) = value;
 }   
}


void msg_window(u8 x1, u8 y1, u8 x2, u8 y2) {
// draw a window with frame and black backgound on the screen

    // window backgound
	FontFill(x1 + 1, y1 + 1, x2 - x1 - 1, y2 - y1 - 1,' ');
	// upper frame
	FontFill(x1 + 1, y1, x2 - x1 - 1, 1,'e');
	// lower frame
	FontFill(x1 + 1, y2, x2 - x1 - 1, 1,'d');
	// left frame
	FontFill(x1 , y1 + 1, 1, y2 - y1 - 1,'c');
	// right frame
	FontFill(x2, y1 + 1, 1 , y2 - y1 - 1,'c');
	// upper, left corner
	PrintChar(x1,y1,'_');
	// upper, right corner
	PrintChar(x2,y1,'`');
	// lower, left corner
	PrintChar(x1,y2,'a');
	// lower, right corner
	PrintChar(x2,y2,'b');
}	



int get_highscore(u8 entry) {
// get the actual highscore from eeprom
  // check the value for entry	
  if (entry > 2) return(0);
	
   // read the eeprom block
  if (!isEepromFormatted() || EepromReadBlock(23, &ebs))
        return(0);   
  return((ebs.data[(entry * 10)+8] * 256) + ebs.data[(entry * 10)+9]);
}



u8 check_highscore(void) {
// check the actual highsore
u8 a;
int i1;
   // read the eeprom block
  if (!isEepromFormatted() || EepromReadBlock(23, &ebs))
        return(0);   
  for(a=0; a<3; a++) {
    i1 = (ebs.data[(a * 10)+8] * 256) + ebs.data[(a * 10)+9];
    if (iSCORE > i1) return(a + 1);
  }

  // highscore is lower as saved highscores 
  return(0);
}



void copy_highsore(u8 entry_from, u8 entry_to) {
// copy a highscore entry to another slot
u8 a;
   // read the eeprom block
  for(a=0; a<10; a++) {
    ebs.data[(entry_to * 10) + a] = ebs.data[(entry_from * 10) + a];
  } 
}


void clear_highsore(u8 entry) {
// clear the name in actual entry and set the score to highscore
u8 a;
  // clear name 
  for(a=0; a<8; a++) {
    ebs.data[(entry * 10) + a] = 0x20;
  }   
  // set score
  ebs.data[(entry * 10) + 8] = iSCORE / 256;
  ebs.data[(entry * 10) + 9] = iSCORE % 256;
}



u8 set_def_highscore(void) {
// write the default highscore list in the EEPROM
  // entry 1
  ebs.data[0] = 'H';
  ebs.data[1] = 'A';
  ebs.data[2] = 'R';
  ebs.data[3] = 'T';
  ebs.data[4] = 'M';
  ebs.data[5] = 'U';
  ebs.data[6] = 'T';
  ebs.data[7] = ' ';
  ebs.data[8] = 0x00;
  ebs.data[9] = 20;	
  // entry 2
  ebs.data[10] = 'M';
  ebs.data[11] = 'A';
  ebs.data[12] = 'R';
  ebs.data[13] = 'I';
  ebs.data[14] = 'E';
  ebs.data[15] = ' ';
  ebs.data[16] = ' ';
  ebs.data[17] = ' ';
  ebs.data[18] = 0x00;
  ebs.data[19] = 15;	
  // entry 3
  ebs.data[20] = 'L';
  ebs.data[21] = 'U';
  ebs.data[22] = 'K';
  ebs.data[23] = 'A';
  ebs.data[24] = 'S';
  ebs.data[25] = ' ';
  ebs.data[26] = ' ';
  ebs.data[27] = ' ';
  ebs.data[28] = 0x00;
  ebs.data[29] = 10;	
  return(EepromWriteBlock(&ebs));
}


u8 view_highscore_entry(u8 x, u8 y, u8 entry, u8 load_data) {
// shows an entry of the higscore
u8 a,c;

  // read the eeprom block
  if (load_data)
  {
    if (!isEepromFormatted() || EepromReadBlock(23, &ebs))
        return(1);   
  }
  entry--;
  for(a = 0; a < 8;a++) {
	c = ebs.data[a + (entry * 10)];  
	PrintChar(x + a, y, c);  
  }
  PrintInt(x + 13, y, (ebs.data[(entry * 10)+8] * 256) + ebs.data[(entry * 10)+9], true);
  return(0);
}



void edit_highscore_entry(u8 entry, u8 cursor_pos, u8 b_mode) {
// edit and view and char in the name of choosed entry    
entry--;
u8 c = ebs.data[(entry * 10) + cursor_pos];
  // proceed up & down button
  if (b_mode == BTN_UP) {
     c++;
     if (c > 'Z') c = ' '; 
     else if (c == '!') c = 'A';
  }
  if (b_mode == BTN_DOWN) {		 
     c--;      
     if (c == 0x1F) c = 'Z';
	 else if (c < 'A') c = ' ';
  }
  ebs.data[(entry * 10) + cursor_pos] = c;

}


void show_highscore_char(u8 entry, u8 position, u8 cursor_on) {
// shows a char of edited name
u8 c = ebs.data[(entry * 10) + position];
    if (cursor_on) PrintChar(14 + position, (entry * 2) + 13, '#');   // show '_'
    else if (c == ' ') PrintChar(14 + position, (entry * 2) + 13, ' ');	// space
    else PrintChar(14 + position, (entry * 2) + 13, c); 	
}





